(defpackage #:vector
  (:use #:common-lisp)
  (:export #:vec3 #:vec4 #:mat4x4
	   #:make-vec3 #:vec3-x #:vec3-y #:vec3-z
	   #:make-vec4 #:vec4-x #:vec4-y #:vec4-z #:vec4-w
	   #:make-matrix #:matrix-i #:matrix-j #:matrix-a
	   #:make-mat4x4 #:mat4x4-a
	   #:sq #:mag #:distance #:normalize #:dot #:cross
	   #:print-vector #:print-matrix
	   #:m* #:m** #:transpose #:inverse
	   #:mat4x4-identity #:trans-from-axes
	   #:rotate-x #:rotate-y #:rotate-z
	   #:translate #:scale #:shear #:reflect-about-plane
	   #:rotate-axis #:screen-to-view #:view-to-screen
	   #:proj-norm #:parrallel #:ortho #:perspective))
	   
	   

(in-package #:vector)

;;; Warning.  Clisp ignores :type in compiled code

;;; With structs defined this way, lisp doesn't automatically create
;;; a type for the struct.
;;; (defstruct (vec3 (:type (vector double-float)))

(declaim (optimize (speed 3) (safety 0)))
(declaim (inline make-vec3) (inline make-vec4))

(defstruct vec3
  (x 0.0d0 :type double-float)
  (y 0.0d0 :type double-float)
  (z 0.0d0 :type double-float))

(defstruct vec4
  (x 0.0d0 :type double-float)
  (y 0.0d0 :type double-float)
  (z 0.0d0 :type double-float)
  (w 0.0d0 :type double-float))

(defstruct matrix
  (i 4 :type fixnum) ;rows
  (j 4 :type fixnum) ;columns
  a)                 ;data

(defstruct mat4x4
  (a (make-array 16 :element-type 'double-float :initial-element 0.0d0) 
     :type (vector double-float 16)))



;(defun sq (x)
;  (declare (double-float x))
;  (* x x))

(defmacro sq (x)
  "Returns the square of X."
  `(let ((val ,x))
     (* val val)))

(defun mag (v)
  "Returns the magnitude of the vector V."
  (typecase v
    (vec3 (sqrt (+ (sq (vec3-x v))
		   (sq (vec3-y v))
		   (sq (vec3-z v)))))
    (vec4 (sqrt (+ (sq (vec4-x v))
		   (sq (vec4-y v))
		   (sq (vec4-z v))
		   (sq (vec4-w v)))))))

(defun distance (a b)
  "Returns the distance between the vectors A and B."
  (typecase a
    (vec3 (sqrt (+ (- (vec3-x a) (vec3-x b))
		   (- (vec3-y a) (vec3-y b))
		   (- (vec3-z a) (vec3-z b)))))
    (vec4 (sqrt (+ (- (vec4-x a) (vec4-x b))
		   (- (vec4-y a) (vec4-y b))
		   (- (vec4-z a) (vec4-z b)))))))

(defun normalize (v)
  "Returns the normalized form of the vector V."
  (typecase v
    (vec3 (let ((magnitude (mag v)))
	    (declare (double-float magnitude))
	    (if (= magnitude 0.0d0)
		(make-vec3 :x 0.0d0 :y 0.0d0 :z 0.0d0)
		(make-vec3 :x (/ (vec3-x v) magnitude)
			   :y (/ (vec3-y v) magnitude)
			   :z (/ (vec3-z v) magnitude)))))
    (vec4 (let ((magnitude (mag v)))
	    (declare (double-float magnitude))
	    (if (= magnitude 0.0d0)
		(make-vec4 :x 0.0d0 :y 0.0d0 :z 0.0d0 :w 0.0d0)
		(make-vec4 :x (/ (vec4-x v) magnitude)
			   :y (/ (vec4-y v) magnitude)
			   :z (/ (vec4-z v) magnitude)
			   :w 0.0d0))))))
#|
;; This little experiment reduced consing but had no significant
;; impact on execution time.
(defvar *normal-vec4* (make-vec4))
(defvar *normal-vec3* (make-vec3))
(let ((magnitude 0.0d0))
  (declare (double-float magnitude))
  
  (defun fast-normalize (v)
    (setf magnitude (mag v))
    (typecase v
      (vec3 (if (= magnitude 0.0d0)
		(progn 
		  (setf (vec3-x *normal-vec3*) 0.0d0)
		  (setf (vec3-y *normal-vec3*) 0.0d0)
		  (setf (vec3-z *normal-vec3*) 0.0d0)
		  *normal-vec3*)
	        (progn 
		  (setf (vec3-x *normal-vec3*) (/ (vec3-x v) magnitude))
		  (setf (vec3-y *normal-vec3*) (/ (vec3-y v) magnitude))
		  (setf (vec3-z *normal-vec3*) (/ (vec3-z v) magnitude))
		  *normal-vec3*)))
      (vec4 (if (= magnitude 0.0d0)
		(make-vec4 :x 0.0d0 :y 0.0d0 :z 0.0d0 :w 0.0d0)
		(progn 
		  (setf (vec4-x *normal-vec4*) (/ (vec4-x v) magnitude))
		  (setf (vec4-y *normal-vec4*) (/ (vec4-y v) magnitude))
		  (setf (vec4-z *normal-vec4*) (/ (vec4-z v) magnitude))
		  (setf (vec4-w *normal-vec4*) 1.0d0)
		  *normal-vec4*)))))
  )
|#
    
(defun dot (u v)
  "Returns the dot product of U and V."
  (typecase u
    (vec3 (+ (* (vec3-x u) (vec3-x v))
	     (* (vec3-y u) (vec3-y v))
	     (* (vec3-z u) (vec3-z v))))
    (vec4 (+ (* (vec4-x u) (vec4-x v))
	     (* (vec4-y u) (vec4-y v))
	     (* (vec4-z u) (vec4-z v))
	     (* (vec4-w u) (vec4-w v))))))

(defun cross (u v)
  "Returns the cross product of U and V."
  (typecase u
    (vec3 (make-vec3 :x (- (* (vec3-y u) (vec3-z v))
			   (* (vec3-z u) (vec3-y v)))
		     :y (- (* (vec3-z u) (vec3-x v))
			   (* (vec3-x u) (vec3-z v)))
		     :z (- (* (vec3-x u) (vec3-y v))
			   (* (vec3-y u) (vec3-x v)))))
    (vec4 (make-vec4 :x (- (* (vec4-y u) (vec4-z v))
			   (* (vec4-z u) (vec4-y v)))
		     :y (- (* (vec4-z u) (vec4-x v))
			   (* (vec4-x u) (vec4-z v)))
		     :z (- (* (vec4-x u) (vec4-y v))
			   (* (vec4-y u) (vec4-x v)))
		     :w 0.0d0))))

(defun print-vector (v)
  (typecase v
    (vec3 (format t "[ ~A ~A ~A ]~%" (vec3-x v) (vec3-y v) (vec3-z v)))
    (vec4 (format t "[ ~A ~A ~A ~A ]~%" (vec4-x v) (vec4-y v) (vec4-z v)
		                        (vec4-w v)))))
(defun vec+ (u v)
  "Adds the vectors U and V."
  (typecase u
    (vec3 (make-vec3 :x (+ (vec3-x u) (vec3-x v))
		     :y (+ (vec3-y u) (vec3-y v))
		     :z (+ (vec3-z u) (vec3-z v))))
    (vec4 (make-vec4 :x (+ (vec4-x u) (vec4-x v))
		     :y (+ (vec4-y u) (vec4-y v))
		     :z (+ (vec4-z u) (vec4-z v))
		     :w (vec4-w u)))))

(defun vec- (u v)
  "Subtracts the vector V from U."
  (typecase u
    (vec3 (make-vec3 :x (- (vec3-x u) (vec3-x v))
		     :y (- (vec3-y u) (vec3-y v))
		     :z (- (vec3-z u) (vec3-z v))))
    (vec4 (make-vec4 :x (- (vec4-x u) (vec4-x v))
		     :y (- (vec4-y u) (vec4-y v))
		     :z (- (vec4-z u) (vec4-z v))
		     :w (vec4-w u)))))

#|
(defmethod m* ((a matrix) (b matrix))
  (if (not (= (matrix-j a) (matrix-i b)))
      (format t "m*: Matrix dimensions incongruous~%")
      ()))

(defmethod m* ((a mat4x4) (b mat4x4))
  )

(defmethod m* ((a vec4) (b mat4x4))
  (make-vec4 :x (+ (* (vec4-x a) (aref (mat4x4-a b) 0))
		   (* (vec4-y a) (aref (mat4x4-a b) 0))
		   (* (vec4-z a) (aref (mat4x4-a b) 0))
		   (* (vec4-w a) (aref (mat4x4-a b) 0)))
	     :y (+ (* (vec4-x a) (aref (mat4x4-a b) 0))
		   (* (vec4-y a) (aref (mat4x4-a b) 0))
		   (* (vec4-z a) (aref (mat4x4-a b) 0))
		   (* (vec4-w a) (aref (mat4x4-a b) 0)))
	     :z (+ (* (vec4-x a) (aref (mat4x4-a b) 0))
		   (* (vec4-y a) (aref (mat4x4-a b) 0))
		   (* (vec4-z a) (aref (mat4x4-a b) 0))
		   (* (vec4-w a) (aref (mat4x4-a b) 0)))
	     :w (+ (* (vec4-x a) (aref (mat4x4-a b) 0))
		   (* (vec4-y a) (aref (mat4x4-a b) 0))
		   (* (vec4-z a) (aref (mat4x4-a b) 0))
		   (* (vec4-w a) (aref (mat4x4-a b) 0)))))
|#

;; Multiplies two 4x4 matrices
;; A and B are the data portions of two mat4x4 structures
;; Returns the data portion for the resulting mat4x4
(defun m*-4x4*4x4 (a b)
  (declare (type (vector double-float 16) a b)) 
  (let ((x (make-array 16 :element-type 'double-float)))
    (setf (aref x 0) (+ (* (aref a 0) (aref b 0))
			(* (aref a 1) (aref b 4))
			(* (aref a 2) (aref b 8))
			(* (aref a 3) (aref b 12))))
    (setf (aref x 1) (+ (* (aref a 0) (aref b 1))
			(* (aref a 1) (aref b 5))
			(* (aref a 2) (aref b 9))
			(* (aref a 3) (aref b 13))))
    (setf (aref x 2) (+ (* (aref a 0) (aref b 2))
			(* (aref a 1) (aref b 6))
			(* (aref a 2) (aref b 10))
			(* (aref a 3) (aref b 14))))
    (setf (aref x 3) (+ (* (aref a 0) (aref b 3))
			(* (aref a 1) (aref b 7))
			(* (aref a 2) (aref b 11))
			(* (aref a 3) (aref b 15))))
    (setf (aref x 4) (+ (* (aref a 4) (aref b 0))
			(* (aref a 5) (aref b 4))
			(* (aref a 6) (aref b 8))
			(* (aref a 7) (aref b 12))))
    (setf (aref x 5) (+ (* (aref a 4) (aref b 1))
			(* (aref a 5) (aref b 5))
			(* (aref a 6) (aref b 9))
			(* (aref a 7) (aref b 13))))
    (setf (aref x 6) (+ (* (aref a 4) (aref b 2))
			(* (aref a 5) (aref b 6))
			(* (aref a 6) (aref b 10))
			(* (aref a 7) (aref b 14))))
    (setf (aref x 7) (+ (* (aref a 4) (aref b 3))
			(* (aref a 5) (aref b 7))
			(* (aref a 6) (aref b 11))
			(* (aref a 7) (aref b 15))))
    (setf (aref x 8) (+ (* (aref a 8) (aref b 0))
			(* (aref a 9) (aref b 4))
			(* (aref a 10) (aref b 8))
			(* (aref a 11) (aref b 12))))
    (setf (aref x 9) (+ (* (aref a 8) (aref b 1))
			(* (aref a 9) (aref b 5))
			(* (aref a 10) (aref b 9))
			(* (aref a 11) (aref b 13))))
    (setf (aref x 10) (+ (* (aref a 8) (aref b 2))
			 (* (aref a 9) (aref b 6))
			 (* (aref a 10) (aref b 10))
			 (* (aref a 11) (aref b 14))))
    (setf (aref x 11) (+ (* (aref a 8) (aref b 3))
			 (* (aref a 9) (aref b 7))
			 (* (aref a 10) (aref b 11))
			 (* (aref a 11) (aref b 15))))
    (setf (aref x 12) (+ (* (aref a 12) (aref b 0))
			 (* (aref a 13) (aref b 4))
			 (* (aref a 14) (aref b 8))
			 (* (aref a 15) (aref b 12))))
    (setf (aref x 13) (+ (* (aref a 12) (aref b 1))
			 (* (aref a 13) (aref b 5))
			 (* (aref a 14) (aref b 9))
			 (* (aref a 15) (aref b 13))))
    (setf (aref x 14) (+ (* (aref a 12) (aref b 2))
			 (* (aref a 13) (aref b 6))
			 (* (aref a 14) (aref b 10))
			 (* (aref a 15) (aref b 14))))
    (setf (aref x 15) (+ (* (aref a 12) (aref b 3))
			 (* (aref a 13) (aref b 7))
			 (* (aref a 14) (aref b 11))
			 (* (aref a 15) (aref b 15))))
    x))

(defun m* (a b)
  (typecase a
    ;; If a is of type vec4, assume b is of type mat4x4
    (vec4 (let ((c (mat4x4-a b)))
	    (make-vec4 :x (+ (* (vec4-x a) (aref c 0))
			     (* (vec4-y a) (aref c 4))
			     (* (vec4-z a) (aref c 8))
			     (* (vec4-w a) (aref c 12)))
		       :y (+ (* (vec4-x a) (aref c 1))
			     (* (vec4-y a) (aref c 5))
			     (* (vec4-z a) (aref c 9))
			     (* (vec4-w a) (aref c 13)))
		       :z (+ (* (vec4-x a) (aref c 2))
			     (* (vec4-y a) (aref c 6))
			     (* (vec4-z a) (aref c 10))
			     (* (vec4-w a) (aref c 14)))
		       :w (+ (* (vec4-x a) (aref c 3))
			     (* (vec4-y a) (aref c 7))
			     (* (vec4-z a) (aref c 11))
			     (* (vec4-w a) (aref c 15))))))
    ;; If a is of type mat4x4, assume b is of type mat4x4 as well
    (mat4x4 (make-mat4x4 :a (m*-4x4*4x4 (mat4x4-a a) (mat4x4-a b))))
    ;; If a is of type matrix, assume b is of type matrix as well
    (matrix nil)
    (t (format t "m*: invalid types~%"))))

;; A user-friendly matrix multiplication function.  Works just like
;; the '*' lisp function.  Since i'm using row vectors, the matrices
;; are applied in a left to right order.
(defun m** (a b &rest rest)
  (do ((m1 a (m* m1 m2))
       (m2 b (pop rest)))
      ((null rest) (m* m1 m2))))

#|
(defun plus (a b &rest rest)
  (do ((m1 a (+ m1 m2))
       (m2 b (pop rest)))
      ((null rest) (+ m1 m2))))
|#

;; Returns the transpose of M  
(defun transpose (m) )

;; Returns the inverse of M
;; M must be of type mat4x4
(defun inverse (m)
  (declare (mat4x4 m))
  (let ((ma (mat4x4-a m))
	(b (make-array 16 :element-type 'double-float))
	(a (make-array 20 :element-type 'double-float)))
    (declare (type (vector double-float 16) ma))
    
    ;; Solve first set of equations
    (setf (aref a 0) (aref ma 0))
    (setf (aref a 1) (aref ma 1))
    (setf (aref a 2) (aref ma 2))
    (setf (aref a 3) (aref ma 3))
    (setf (aref a 4) 1.0d0)

    (setf (aref a 5) (aref ma 4))
    (setf (aref a 6) (aref ma 5))
    (setf (aref a 7) (aref ma 6))
    (setf (aref a 8) (aref ma 7))
    (setf (aref a 9) 0.0d0)

    (setf (aref a 10) (aref ma 8))
    (setf (aref a 11) (aref ma 9))
    (setf (aref a 12) (aref ma 10))
    (setf (aref a 13) (aref ma 11))
    (setf (aref a 14) 0.0d0)

    (setf (aref a 15) (aref ma 12))
    (setf (aref a 16) (aref ma 13))
    (setf (aref a 17) (aref ma 14))
    (setf (aref a 18) (aref ma 15))
    (setf (aref a 19) 0.0d0)

    (multiple-value-bind (e f g h) (gauss-elim a 4)
      (if (null e)
	  (return-from inverse nil)
	  (block assign-values
	    (setf (aref b 0)  e)
	    (setf (aref b 4)  f)
	    (setf (aref b 8)  g)
	    (setf (aref b 12) h))))

    ;; Solve second set of equations
    (setf (aref a 0) (aref ma 0))
    (setf (aref a 1) (aref ma 1))
    (setf (aref a 2) (aref ma 2))
    (setf (aref a 3) (aref ma 3))
    (setf (aref a 4) 0.0d0)

    (setf (aref a 5) (aref ma 4))
    (setf (aref a 6) (aref ma 5))
    (setf (aref a 7) (aref ma 6))
    (setf (aref a 8) (aref ma 7))
    (setf (aref a 9) 1.0d0)

    (setf (aref a 10) (aref ma 8))
    (setf (aref a 11) (aref ma 9))
    (setf (aref a 12) (aref ma 10))
    (setf (aref a 13) (aref ma 11))
    (setf (aref a 14) 0.0d0)

    (setf (aref a 15) (aref ma 12))
    (setf (aref a 16) (aref ma 13))
    (setf (aref a 17) (aref ma 14))
    (setf (aref a 18) (aref ma 15))
    (setf (aref a 19) 0.0d0)

    (multiple-value-bind (e f g h) (gauss-elim a 4)
      (if (null e)
	  (return-from inverse nil)
	  (block assign-values
	    (setf (aref b 1)  e)
	    (setf (aref b 5)  f)
	    (setf (aref b 9)  g)
	    (setf (aref b 13) h))))

    ;; Solve third set of equations
    (setf (aref a 0) (aref ma 0))
    (setf (aref a 1) (aref ma 1))
    (setf (aref a 2) (aref ma 2))
    (setf (aref a 3) (aref ma 3))
    (setf (aref a 4) 0.0d0)

    (setf (aref a 5) (aref ma 4))
    (setf (aref a 6) (aref ma 5))
    (setf (aref a 7) (aref ma 6))
    (setf (aref a 8) (aref ma 7))
    (setf (aref a 9) 0.0d0)

    (setf (aref a 10) (aref ma 8))
    (setf (aref a 11) (aref ma 9))
    (setf (aref a 12) (aref ma 10))
    (setf (aref a 13) (aref ma 11))
    (setf (aref a 14) 1.0d0)

    (setf (aref a 15) (aref ma 12))
    (setf (aref a 16) (aref ma 13))
    (setf (aref a 17) (aref ma 14))
    (setf (aref a 18) (aref ma 15))
    (setf (aref a 19) 0.0d0)

    (multiple-value-bind (e f g h) (gauss-elim a 4)
      (if (null e)
	  (return-from inverse nil)
	  (block assign-values
	    (setf (aref b 2)  e)
	    (setf (aref b 6)  f)
	    (setf (aref b 10) g)
	    (setf (aref b 14) h))))

    ;; Solve fourth set of equations
    (setf (aref a 0) (aref ma 0))
    (setf (aref a 1) (aref ma 1))
    (setf (aref a 2) (aref ma 2))
    (setf (aref a 3) (aref ma 3))
    (setf (aref a 4) 0.0d0)

    (setf (aref a 5) (aref ma 4))
    (setf (aref a 6) (aref ma 5))
    (setf (aref a 7) (aref ma 6))
    (setf (aref a 8) (aref ma 7))
    (setf (aref a 9) 0.0d0)

    (setf (aref a 10) (aref ma 8))
    (setf (aref a 11) (aref ma 9))
    (setf (aref a 12) (aref ma 10))
    (setf (aref a 13) (aref ma 11))
    (setf (aref a 14) 0.0d0)

    (setf (aref a 15) (aref ma 12))
    (setf (aref a 16) (aref ma 13))
    (setf (aref a 17) (aref ma 14))
    (setf (aref a 18) (aref ma 15))
    (setf (aref a 19) 1.0d0)

    (multiple-value-bind (e f g h) (gauss-elim a 4)
      (if (null e)
	  (return-from inverse nil)
	  (block assign-values
	    (setf (aref b 3)  e)
	    (setf (aref b 7)  f)
	    (setf (aref b 11) g)
	    (setf (aref b 15) h))))

    (make-mat4x4 :a b)))

;; Helper function for GAUSS-ELIM
(defun get-p (a i w n)
  (do ((p i (+ p 1)))
      ((= p n) p)
    (if (not (= 0.0d0 (aref a (+ (* p w) i)))) (return-from get-p p))))

;; Perform gaussian elimination on a system of equations
;; Returns the 4 unknowns in the system (hardcoded size for now)
;; A is an array of coefficients for the system of equations
;; N is the number of coefficients on the left side of the equation  	  
(defun gauss-elim (a n)
  (let ((x (make-array 4 :initial-element 0.0d0 :element-type 'double-float))
	(w (+ n 1)))
    (dotimes (i (- n 1))
      (let ((p (get-p a i w n)))
	;(format t "p = ~A ~%" p)
	(if (= p n) (return-from gauss-elim nil))
	(if (not (= p i)) (dotimes (k w)
			    (let ((temp (aref a (+ (* i w) k))))
			      (setf (aref a (+ (* i w) k)) 
				    (aref a (+ (* p w) k)))
			      (setf (aref a (+ (* p w) k)) temp))))
	(do ((j (+ i 1) (+ j 1)))
	    ((= j n))
	    (let ((m (/ (aref a (+ (* j w) i))
			(aref a (+ (* i w) i)))))
	      (do ((k 0 (+ k 1)))
		  ((> k n))
		  (setf (aref a (+ (* j w) k)) (- (aref a (+ (* j w) k))
						  (* m (aref a (+ (* i w) k))))))))))
    
    (if (= (aref a (+ (* (- n 1) 
			 w) 
		      (- n 1))) 
	   0.0d0) 
	(return-from gauss-elim nil))

    (setf (aref x (- n 1)) (/ (aref a (+ (* (- n 1) w)
					 n))
			      (aref a (+ (* (- n 1) w)
					 (- n 1)))))

    (do ((i (- n 2) (- i 1))
	 (s 0.0d0 0.0d0))
	((< i 0))
	(do ((j (+ i 1) (+ j 1)))
	    ((= j n))
	    (setf s (+ s (* (aref a (+ (* i w) j))
			    (aref x j)))))
	(setf (aref x i) (/ (- (aref a (+ (* i w) n)) 
			       s)
			    (aref a (+ (* i w) i)))))

    ;(format t "GAUSS-ELIM: made it to the end~%")
    (values (aref x 0) (aref x 1) (aref x 2) (aref x 3))))

(defun print-matrix (m)
  (typecase m
    (mat4x4 (print-mat4x4 m))
    (vec3 (print-vector m))
    (vec4 (print-vector m))
    (matrix nil)))

(defun print-mat4x4 (mat)
  (let ((m (mat4x4-a mat)))
    (format t "[ ~A ~A ~A ~A ]~%" (aref m 0) (aref m 1)
	    (aref m 2) (aref m 3))
    (format t "[ ~A ~A ~A ~A ]~%" (aref m 4) (aref m 5)
	    (aref m 6) (aref m 7))
    (format t "[ ~A ~A ~A ~A ]~%" (aref m 8) (aref m 9)
	    (aref m 10) (aref m 11))
    (format t "[ ~A ~A ~A ~A ]~%" (aref m 12) (aref m 13)
	    (aref m 14) (aref m 15))))

(defun mat4x4-identity ()
  (let* ((m (make-mat4x4))
	 (a (mat4x4-a m)))
    (setf (aref a 0) 1.0d0)
    (setf (aref a 5) 1.0d0)
    (setf (aref a 10) 1.0d0)
    (setf (aref a 15) 1.0d0)
    m))

(defun trans-from-axes (x y z)
  "Constructs a 4x4 transformation matrix from the coordinate axes
  of the desired coordinate system."
  (declare (vec4 x y z))
  (let* ((m (make-mat4x4))
	 (a (mat4x4-a m)))
    (setf (aref a 0)  (vec4-x x))
    (setf (aref a 1)  (vec4-y x))
    (setf (aref a 2)  (vec4-z x))
    (setf (aref a 3)  0.0d0)
    (setf (aref a 4)  (vec4-x y))
    (setf (aref a 5)  (vec4-y y))
    (setf (aref a 6)  (vec4-z y))
    (setf (aref a 7)  0.0d0)
    (setf (aref a 8)  (vec4-x z))
    (setf (aref a 9)  (vec4-y z))
    (setf (aref a 10) (vec4-z z))
    (setf (aref a 11) 0.0d0)
    (setf (aref a 12) 0.0d0)
    (setf (aref a 13) 0.0d0)
    (setf (aref a 14) 0.0d0)
    (setf (aref a 15) 1.0d0)
    m))

;; Return a matrix which:
;; Rotates DEG degrees about the x axis.
(defun rotate-x (deg)
  (let* ((m (make-mat4x4))
	 (a (mat4x4-a m))
	 (rad (/ (* deg pi) 180.0d0)))
    (declare (double-float rad))
    (setf (aref a 0) 1.0d0)
    (setf (aref a 5) (cos rad))
    (setf (aref a 6) (sin rad))
    (setf (aref a 9) (- (sin rad)))
    (setf (aref a 10) (cos rad))
    (setf (aref a 15) 1.0d0)
    m))

;; Return a matrix which:   
;; Rotates DEG degrees about the y axis. 
(defun rotate-y (deg)
  (let* ((m (make-mat4x4))
	 (a (mat4x4-a m))
	 (rad (/ (* deg pi) 180.0d0)))
    (declare (double-float rad))
    (setf (aref a 0) (cos rad))
    (setf (aref a 2) (- (sin rad)))
    (setf (aref a 5) 1.0d0)
    (setf (aref a 8) (sin rad))
    (setf (aref a 10) (cos rad))
    (setf (aref a 15) 1.0d0)
    m))

;; Return a matrix which:
;; Rotates DEG degrees about the z axis.
(defun rotate-z (deg)
  (let* ((m (make-mat4x4))
	 (a (mat4x4-a m))
	 (rad (/ (* deg pi) 180.0d0)))
    (declare (double-float rad))
    (setf (aref a 0) (cos rad))
    (setf (aref a 1) (sin rad))
    (setf (aref a 4) (- (sin rad)))
    (setf (aref a 5) (cos rad))
    (setf (aref a 10) 1.0d0)
    (setf (aref a 15) 1.0d0)
    m))

;; Return a matrix which:
;; Performs a translation.
(defun translate (x y z)
  (let* ((m (make-mat4x4))
	 (a (mat4x4-a m)))
    (setf (aref a 0)   1.0d0)
    (setf (aref a 5)   1.0d0)
    (setf (aref a 10)  1.0d0)    
    (setf (aref a 12) (float x 1.0d0))
    (setf (aref a 13) (float y 1.0d0))
    (setf (aref a 14) (float z 1.0d0))
    (setf (aref a 15) 1.0d0)
    m))

;; Return a matrix which:
;; Performs a three dimensional scaling. 
(defun scale (x y z)
  (let* ((m (make-mat4x4))
	 (a (mat4x4-a m)))
    (setf (aref a 0) (float x 1.0d0))
    (setf (aref a 5) (float y 1.0d0))
    (setf (aref a 10) (float z 1.0d0))
    (setf (aref a 15) 1.0d0)
    m))

;; Return a matrix which:
;; Performs a three dimensional shearing
(defun shear (&key (y-by-x 0.0d0) (z-by-x 0.0d0) (x-by-y 0.0d0) (z-by-y 0.0d0)
		   (x-by-z 0.0d0) (y-by-z 0.0d0))
  (let* ((m (make-mat4x4))
	 (a (mat4x4-a m)))
    (setf (aref a 0) 1.0d0)
    (setf (aref a 5) 1.0d0)
    (setf (aref a 10) 1.0d0)
    (setf (aref a 15) 1.0d0)
    (setf (aref a 1) y-by-x)
    (setf (aref a 2) z-by-x)
    (setf (aref a 4) x-by-y)
    (setf (aref a 6) z-by-y)
    (setf (aref a 8) x-by-z)
    (setf (aref a 9) y-by-z)    
    m))

;; Return a matrix which:
(defun reflect-about-plane (plane)
  nil)

;; Return a matrix which:
;; Rotates DEG degrees about AXIS.
(defun rotate-axis (deg axis)
  (declare (vec4 axis))
  nil)

;; Return
(defun viewport ()
  nil)

;; Returns the x and y coordinates of a screen pixel in the canonical view volume
;; i is x-value(column) of the pixel
;; j is y-value(row) of the pixel
;; maxi is the maximum i-value for a pixel
;; maxj is the maximum j-value for a pixel
(defun screen-to-view (i j maxi maxj)
  (let ((left -1.0d0)
	(top   1.0d0))
    (values (+ left (* (/ i maxi) 2.0d0))
	    (- top  (* (/ j maxj) 2.0d0)))))

(defun fast-screen-to-view (i j maxi maxj)
  (declare (double-float i j maxi maxj))
  (let ((left -1.0d0)
	(top   1.0d0))
    (values (+ left (* (/ i maxi) 2.0d0))
	    (- top  (* (/ j maxj) 2.0d0)))))

;; (X,Y) are the view space coordinates
;; MAXI and MAXJ are the maximum (I,J) pixel coordinates
(defun view-to-screen (x y maxi maxj)
  (values (truncate (* (/ (+ x 1.0d0) 2.0d0) maxi))
	  (truncate (* (/ (- y 1.0d0) -2.0d0) maxj))))

;; Return a matrix which:
;; Transforms a "roughly cubic" viewing volume into the canonical
;; view volume.  The arguments can be thought of as three pairs
;; of parrallel planes which form the volume to be transformed.
(defun proj-norm (xmin xmax ymin ymax zmin zmax)
  (declare (double-float xmin xmax ymin ymax zmin zmax))
  (let* ((m (make-mat4x4))
	 (a (mat4x4-a m)))
    (setf (aref a 0) (/ 2 (- xmax xmin)))
    (setf (aref a 5) (/ 2 (- ymax ymin)))
    (setf (aref a 10) (/ 2 (- zmax zmin)))
    (setf (aref a 12) (- (/ (+ xmax xmin)
			    (- xmax xmin))))
    (setf (aref a 13) (- (/ (+ ymax ymin)
			    (- ymax ymin))))
    (setf (aref a 14) (- (/ (+ zmax zmin)
			    (- zmax zmin))))
    (setf (aref a 15) 1.0d0)
    m))

#|
(defun proj-norm (xmin xmax ymin ymax zmin zmax)
  (declare (double-float xmin xmax ymin ymax zmin zmax))
  (m* (scale (/ 2 (- xmax xmin))
	     (/ 2 (- ymax ymin))
	     (/ 2 (- zmax zmin)))
      (translate (- (/ (+ xmax xmin) (- xmax xmin)))
		 (- (/ (+ ymax ymin) (- ymax ymin)))
		 (- (/ (+ zmax zmin) (- zmax zmin))))))
|#

;; Return a matrix which:
;; Does a simple parrallel projection onto the z=0 plane.
(defun parrallel ()
  (scale 1.0d0 1.0d0 0.0d0))

;; Return a matrix which:
;; Performs a general orthographic projection.
(defun ortho (xmin xmax ymin ymax zmin zmax)
  (m* (proj-norm xmin xmax ymin ymax zmin zmax)
      (parrallel)))

;; Return a matrix which:
;; Performs a perspective projection.
(defun perspective ()
  nil)

;; Destructively performs a perspective divide on a vector.
;; Returns the modified vector.
(defun pdiv! (v)
  (declare (vec4 v))
  (let ((w (vec4-w v)))
    (setf (vec4-x v) (/ (vec4-x v) w))
    (setf (vec4-y v) (/ (vec4-y v) w))
    (setf (vec4-z v) (/ (vec4-z v) w))
    (setf (vec4-w v) (/ (vec4-w v) w)))
  v)

