(in-package #:lascivious)

(use-package "ALIEN")
(use-package "C-CALL")

(declaim (optimize (speed 3) (safety 0)))

(defun max-vol-value (vol)
  "Returns the maximum voxel value in the volume VOL."
  (let ((size (* vol-width vol-height vol-images))
	(max 0))
    (dotimes (i size max)
      (if (< max (deref rawdata i))
	  (setf max (deref rawdata i)))
      (if (= (mod i 1000) 0) 
	  (format t "i = ~A, max = ~A~%" i max)))))

(defun write-pgm (path data x y)
  "Writes a PGM file named PATH.
  DATA is a 2d array of bytes.
  X and Y are the dimensions of the image."
  (with-open-file (p path :direction :output)
    (format p "P2 ~A ~A 255" x y)
    (dotimes (j y)
      (dotimes (i x) 
	(print (aref data j i) p)))))

(defun write-ppm (path data x y)
  "Writes a PPM file named PATH.
  DATA is a 2d array of bytes.
  X and Y are the dimensions of the image."
  (with-open-file (p path :direction :output :element-type 'unsigned-byte)
    (format p "P6 ~A ~A 255~%" x y)
    (dotimes (j y)
      (dotimes (i x)
	(let ((pixel-value (aref data j i)))
	  (write-byte (pv-r pixel-value) p)
	  (write-byte (pv-g pixel-value) p)
	  (write-byte (pv-b pixel-value) p))))))

;; CORNERS is a list containing the volume corner coordinates in view space
(defun get-xybounds (corners)
  (let ((minx -1.0d0)
	(maxx  1.0d0)
	(miny -1.0d0)
	(maxy  1.0d0))
    (dolist (point corners (values minx maxx miny maxy))
      (if (< (vec4-x point) minx) (setf minx (vec4-x point)))
      (if (> (vec4-x point) maxx) (setf maxx (vec4-x point)))
      (if (< (vec4-y point) miny) (setf miny (vec4-y point)))	    
      (if (> (vec4-y point) maxy) (setf maxy (vec4-y point))))))

;; CORNERS is a list containing the volume corner coordinates in view space
(defun get-ijbounds (vol corners)
  "Returns the bounds of the volume VOL in screen space.  CORNERS is a list
   of points corresponding to the VOL's corner coordinates in view space." 
  (let ((vol-max-width (volume-vol-max-width vol))
	(vol-max-height (volume-vol-max-height vol)))
    (multiple-value-bind (minx maxx miny maxy) (get-xybounds corners)
      (multiple-value-bind (mini minj) (view-to-screen minx 
						       maxy 
						       vol-max-width 
						       vol-max-height)
	(multiple-value-bind (maxi maxj) (view-to-screen maxx 
							 miny 
							 vol-max-width
							 vol-max-height)
	  (values mini maxi minj maxj))))))

;; This function is used to find the bounding box for a volume in world space.
;; The aforementioned bounding box can then be used to calculate the
;;   projection transformation, among other things.
;; CORNERS is a list containing the volume corner coordinates in world space
(defun get-projnorm-bounds (vol corners)
  (let ((minx 0.0d0)
	(maxx (volume-vol-max-width vol))
	(miny 0.0d0)
	(maxy (volume-vol-max-height vol))
	(minz 0.0d0)
	(maxz (volume-vol-max-depth vol)))
    (dolist (point corners (values minx maxx miny maxy minz maxz))
      (if (< (vec4-x point) minx) (setf minx (vec4-x point)))
      (if (> (vec4-x point) maxx) (setf maxx (vec4-x point)))
      (if (< (vec4-y point) miny) (setf miny (vec4-y point)))	    
      (if (> (vec4-y point) maxy) (setf maxy (vec4-y point)))
      (if (< (vec4-z point) minz) (setf minz (vec4-z point)))	    
      (if (> (vec4-z point) maxz) (setf maxz (vec4-z point))))))

;; Returns TRUE iff (X,Y,Z) lies within the roughly cubic volume
;; defined by BOUNDS.
;; BOUNDS is ...
(defun point-within-bounds? (x y z bounds)
  t)

(defun nearest-neighbor (x y z)
  "Returns the VOX-COORD corresponding to the nearest neighbor of
  the volume coordinate (X,Y,Z)."
  (declare (double-float x y z))
  (make-vox-coord :x (truncate (+ x 0.5d0))
		  :y (truncate (+ y 0.5d0))
		  :z (truncate (+ z 0.5d0))))

(defun nearest-neighbor-val (vol x y z)
  "Returns the value of the nearest neighbor to the volume coordinate
  (X,Y,Z)."
  (declare (double-float x y z))
  (deref rawdata (+ (* (truncate (+ z 0.5d0)) vol-width vol-height)
		    (* (truncate (+ y 0.5d0)) vol-width)
		    (truncate (+ x 0.5d0)))))

;; Retrieves a voxel value from RAWDATA at the specified voxel coordinate
(defvar vol-width-fixnum)
(defvar vol-height-fixnum)
(declaim (type fixnum vol-width-fixnum vol-height-fixnum))
(declaim (inline get-vol-value))
(defun get-vol-value (vol x y z)
  "Returns the voxel value of the volume coordinate (X,Y,Z)."
  (declare (fixnum x y z))
    (deref rawdata (+ (the fixnum (* (the fixnum (* z vol-width-fixnum)) vol-height-fixnum))
		      (the fixnum (* y vol-width-fixnum))
		      x)))

;(defmacro get-vol-value (vol x y z)
;  `(deref rawdata (+ (the fixnum (* (the fixnum (* ,z vol-width-fixnum)) vol-height-fixnum))
;		     (the fixnum (* ,y vol-width-fixnum))
;		      ,x)))
#|
(defvar *rawdata* (make-array '(256 256 110) :element-type '(unsigned-byte 8)))
(defun convert-rawdata (vol)
  (dotimes (x 256)
    (dotimes (y 256)
      (dotimes (z 110)
	(setf (aref *rawdata* x y z) (get-vol-value vol x y z))))))

(defmacro fast-get-vol-value (vol x y z)
  `(the (unsigned-byte 8) (aref *rawdata* ,x ,y ,z)))

(defmacro fast-get-vol-value (vol x y z)
  `(the (unsigned-byte 8) (deref rawdata ,x ,y ,z)))
|#

;(declaim (inline get-decimal))
;(defun get-decimal (x)
;  (declare (double-float x))
;  (- x (the double-float (ffloor x))))

(defmacro get-decimal (x)
  "Returns the decimal part of X."
  `(- ,x (the double-float (ffloor ,x))))

;(declaim (inline linear-interpolate))
;(defun linear-interpolate (val1 val2 pos)
;  (declare (double-float pos))
;  (+ val1
;     (- (* pos
;	     (- val1 val2)))))

(defmacro linear-interpolate (val1 val2 pos)
  "Linearly interpolates between VAL1 and VAL2, and
  returns the value."
  `(+ ,val1
      (- (* ,pos
	    (- ,val1 ,val2)))))
  
(defun trilinear-interpolate (vol x y z)
  "Determines the voxel value of the volume sample taken at
  (X,Y,Z) by trilinear interpolation and returns that value."
  (declare (double-float x y z))
  
  (if (= (get-decimal x) 0.0d0) (setf x (+ x 0.001d0)))
  (if (= (get-decimal y) 0.0d0) (setf y (+ y 0.001d0)))
  (if (= (get-decimal z) 0.0d0) (setf z (+ z 0.001d0)))  
  
  (let ((xdec (get-decimal x))
	(ydec (get-decimal y))
	(tx (the (unsigned-byte 32) (truncate x)))
	(ty (the (unsigned-byte 32) (truncate y)))
	(tz (the (unsigned-byte 32) (truncate z)))
	(cx (the (unsigned-byte 32) (ceiling x)))
	(cy (the (unsigned-byte 32) (ceiling y)))
	(cz (the (unsigned-byte 32) (ceiling z)))
	(x0 0.0d0)
	(x1 0.0d0)
	(x2 0.0d0)
	(x3 0.0d0)
	(y0 0.0d0)
	(y1 0.0d0))
    (declare (double-float xdec ydec x0 x1 x2 x3 y0 y1))
    
    (setf x0 (linear-interpolate (get-vol-value vol tx cy  tz)
				 (get-vol-value vol cx  cy  tz) 
				 xdec))
    (setf x1 (linear-interpolate (get-vol-value vol tx cy  cz)
				 (get-vol-value vol cx  cy  cz)
				 xdec))
    (setf x2 (linear-interpolate (get-vol-value vol tx ty tz)
				 (get-vol-value vol cx  ty tz)
				 xdec))
    (setf x3 (linear-interpolate (get-vol-value vol tx ty cz)
				 (get-vol-value vol cx  ty cz)
				 xdec))

    (setf y0 (linear-interpolate x0 x2 ydec))
    (setf y1 (linear-interpolate x1 x3 ydec))

    (truncate (linear-interpolate y0 y1 (get-decimal z)))))

(defun max-intensity(curmax new)
  "Return the larger of CURMAX and NEW."
  (if (> new curmax)
      new
      curmax))

(defun central-diff (vol x y z)
  "Returns the gradient(a VEC3) of the sample point (X,Y,Z) in
  the volume using the central difference operator."
  (declare (fixnum x y z))
  (make-vec3 :x (float (- (get-vol-value vol  (- x 1) y z)
			  (get-vol-value vol (+ x 1) y z))
		       1.0d0)
	     :y (float (- (get-vol-value vol x (- y 1) z)
			  (get-vol-value vol x (+ y 1) z))
		       1.0d0)
	     :z (float (- (get-vol-value vol x y (- z 1))
			  (get-vol-value vol x y (+ z 1)))
		       1.0d0)))

(defun calc-gradient (vol p)
  "Calls CENTRAL-DIFF on the sample point P and returns the value."
  (declare (vox-coord p))
  (central-diff vol (vox-coord-x p) (vox-coord-y p) (vox-coord-z p)))

(defun get-gradient (vol p)
  "Returns a voxel P's gradient by retrieving it from the gradient
  look-up-table in VOL."
  (declare (vox-coord p))
  (aref (volume-grad-lut vol) (vox-coord-x p) (vox-coord-y p) (vox-coord-z p)))

(defun classify-mat (vol val)
  "Returns the material corresponding to the scalar voxel value VAL."
  (aref (volume-materials vol)
	(aref (volume-mat-lut vol) val)))

;(declaim (inline classify))
(defun classify (vol val grad)
  "Returns the opacity, given the voxel value and the voxel's gradient."
  (* (aref (volume-scalar-opacity-lut vol) val)
     (aref (volume-grad-opacity-lut vol) (truncate (/ (mag grad)
						      2)))))

;; Returns the pixel value of the sampled point.
;; This includes performing lighting calculations to determine the color
;; and the opacity obtained from classification.
(defun shade (context mat gradient opacity)
  ;(declare (single-float opacity))
  
  (let ((light0 (aref (render-context-lights context) 0))
	(dotprod (dot gradient
		      (lightsource-dir (aref (render-context-lights context) 0)))))
    (if (< dotprod 0.0d0) (setf dotprod 0.0d0)) 
    (make-pv :r (+ (the fixnum (truncate (* (material-amb-r mat)
					    (lightsource-r light0))))
		   (the fixnum (truncate (* (material-dif-r mat)
					    (lightsource-r light0)
					    dotprod))))
	     :g (+ (the fixnum (truncate (* (material-amb-g mat)
					    (lightsource-g light0))))
		   (the fixnum (truncate (* (material-dif-g mat)
					    (lightsource-g light0)
					    dotprod))))
	     :b (+ (the fixnum (truncate (* (material-amb-b mat)
					    (lightsource-b light0))))
		   (the fixnum (truncate (* (material-dif-b mat)
					    (lightsource-b light0)
					    dotprod))))
	     :o opacity)))

;; Combines the accumulated pixel value with the new sample point's
;; color.
(defun accumulate (acc val)
  "Combines the pixel values ACC and VAL, and returns the result."
  (declare (pv acc val))
  (let ((trans (- 1.0 (pv-o acc))))
    (declare (single-float trans))
    (make-pv :r (+ (pv-r acc)
		   (the fixnum (truncate (* trans
					    (* (pv-r val)
					       (pv-o val))))))
	     :g (+ (pv-g acc)
		   (the fixnum (truncate (* trans
					    (* (pv-g val)
					       (pv-o val))))))
	     :b (+ (pv-b acc)
		   (the fixnum (truncate (* trans
					    (* (pv-b val)
					       (pv-o val))))))
	     :o (+ (pv-o acc)
		   (* trans (pv-o val))))))

;; A simplistic composite function for use with
;;   maximum intensity projections.
;; Color is assumed to be greyscale and there is no shading.
;; Color is also assumed to be the same intensity as the voxel value.
;; ACC is the accumulated pixel value.
;; (X,Y,Z) is the sample point.
#|
(defun composite-simple (context acc x y z)
  (declare (pv acc) (double-float x y z))
  (let ((max (max-intensity (pv-r acc) (nearest-neighbor-val context x y z))))
    (make-pv :r max :g max :b max :o (/ max 255.0))))
|#

;; A better composite function which will handle classification and shading.
(defun composite (vol context acc x y z)
  (declare (pv acc) (double-float x y z))
  (let ((gradient (calc-gradient vol (nearest-neighbor x y z)))
	(value (trilinear-interpolate vol x y z)))
    (accumulate acc (shade context
		           (classify-mat vol value)
			   (normalize gradient)
			   (classify vol value gradient)))))

(defun composite-general (vol context acc x y z)
  (declare (pv acc) (double-float x y z))
  (let ((gradient (funcall (volume-calc-grad-func vol) vol (nearest-neighbor x y z)))
	(value (funcall (volume-interpolate-func vol) vol x y z)))
    (accumulate acc (funcall (volume-shade-func vol)
			     context
			     (classify-mat vol value)
			     (normalize gradient)
			     (classify vol value gradient)))))

(defun cast-ray (vol context m x y dx dy dz entry-point exit-point)
  (declare (volume vol) (render-context context) (mat4x4 m) (double-float x y dx dy dz)
	   (vec4 entry-point exit-point))
  ;(format t "Ray cast from (~A, ~A)~%" x y)

  (let ((principle-axis-func #'vec4-z)
	(sense-func #'<))
    (let ((pixel-value (make-pv))
	  (exit-val (funcall principle-axis-func exit-point))
	  (vmw (- (volume-vol-max-width vol) 1.0d0))
	  (vmh (- (volume-vol-max-height vol) 1.0d0))
	  (vmd (- (volume-vol-max-depth vol) 1.0d0)))
      
      (do* ((vpoint (make-vec4 :x x :y y :z 1.0d0 :w 1.0d0))
	    (opoint (m* vpoint m)))
	  ((or (> (pv-o pixel-value) 0.97)
	       (funcall sense-func (funcall principle-axis-func opoint) exit-val)) pixel-value)
      
        ;(format t "   Object coordinate (~A, ~A, ~A)~%" 
        ;       (vec4-x opoint) (vec4-y opoint) (vec4-z opoint))

        ; Avoid the edges of the data set because I don't bounds check
        ; when interpolating the sample point.  This is a problem when
        ; the sample point has a coordinate of zero, for instance.
        ; I should probably fix this at some point.
	(if (and (>= (vec4-x opoint) 1.0d0)
		 (<= (vec4-x opoint) vmw)
		 (>= (vec4-y opoint) 1.0d0)
		 (<= (vec4-y opoint) vmh)
		 (>= (vec4-z opoint) 1.0d0)
		 (<= (vec4-z opoint) vmd))
	    (setf pixel-value (composite-general vol context
						 pixel-value 
						 (vec4-x opoint) 
						 (vec4-y opoint) 
						 (vec4-z opoint))))

	(setf (vec4-x opoint) (+ (vec4-x opoint) dx))
	(setf (vec4-y opoint) (+ (vec4-y opoint) dy))
	(setf (vec4-z opoint) (+ (vec4-z opoint) dz))))))

(defun cast-ray-slow (vol context m x y unused1 unused2 unused3 dz)
  (declare (volume vol) (render-context context) (mat4x4 m) (double-float x y dz))
  ;(format t "Ray cast from (~A, ~A)~%" x y)
  
  (let ((pixel-value (make-pv))
	(vmw (- (volume-vol-max-width vol) 1.0d0))
	(vmh (- (volume-vol-max-height vol) 1.0d0))
	(vmd (- (volume-vol-max-depth vol) 1.0d0)))
    ;(declare (double-float vmw vmh vmd))
    (do* ((vpoint (make-vec4 :x x :y y :z 1.0d0 :w 1.0d0))
	  (opoint (m* vpoint m) (m* vpoint m)))
	 ((or (> (pv-o pixel-value) 0.95)
	      (< (vec4-z vpoint) -1.0d0)) pixel-value)

      ; Avoid the edges of the data set because I don't bounds check
      ; when interpolating the sample point.  This is a problem when
      ; the sample point has a coordinate of zero, for instance.
      ; I should probably fix this at some point.
      (if (and (>= (vec4-x opoint) 1.0d0)
	       (<= (vec4-x opoint) vmw)
	       (>= (vec4-y opoint) 1.0d0)
	       (<= (vec4-y opoint) vmh)
	       (>= (vec4-z opoint) 1.0d0)
	       (<= (vec4-z opoint) vmd))
	  (setf pixel-value (composite-general vol context
					       pixel-value 
					       (vec4-x opoint) 
					       (vec4-y opoint) 
					       (vec4-z opoint))))
      
      
      (setf (vec4-z vpoint) (- (vec4-z vpoint) dz)))))

;; Not feasible at this time
(defun compute-gradients (vol)
  (declare (volume vol))
  (setf (volume-grad-lut vol)
	(make-array '(256 256 110)
		    :element-type 'vec3))

  (let ((vmw (truncate (volume-vol-max-width vol)))
	(vmh (truncate (volume-vol-max-height vol)))
	(vmd (truncate (volume-vol-max-depth vol))))
    (dotimes (x (volume-vol-width vol))
      (dotimes (y (volume-vol-height vol))
	(dotimes (z (volume-vol-images vol))
	  (if (and (> x 0) (> y 0) (> z 0) (< x vmw) (< y vmh) (< z vmd))
	      (setf (aref (volume-grad-lut vol) x y z)
		    (central-diff vol x y z))))))))

(defun make-quads-from-corners (corners)
  (let ((quad-array (make-array 6 :type (simple-array quad (*)))))
    ; 0 - Top
    ; 1 - Bottom
    ; 2 - Front
    ; 3 - Back
    ; 4 - Left
    ; 5 - Right
    (setf (aref 0 quad-array) (make-quad :a (elt corners 4) :b (elt corners 5)
					 :c (elt corners 6) :d (elt corners 7)
					 :n (normalize
					     (cross (vec- (elt corners 5) (elt corners 4))
						    (vec- (elt corners 6) (elt corners 4))))))
    (setf (aref 1 quad-array) (make-quad :a (elt corners 0) :b (elt corners 1)
					 :c (elt corners 2) :d (elt corners 3)
					 :n (normalize
					     (cross (vec- (elt corners 1) (elt corners 0))
						    (vec- (elt corners 2) (elt corners 0))))))
    (setf (aref 2 quad-array) (make-quad :a (elt corners 7) :b (elt corners 6)
					 :c (elt corners 2) :d (elt corners 3)
					 :n (normalize
					     (cross (vec- (elt corners 6) (elt corners 7))
						    (vec- (elt corners 2) (elt corners 7))))))
    (setf (aref 3 quad-array) (make-quad :a (elt corners 0) :b (elt corners 1)
					 :c (elt corners 5) :d (elt corners 4)
					 :n (normalize
					     (cross (vec- (elt corners 1) (elt corners 0))
						    (vec- (elt corners 5) (elt corners 0))))))
    (setf (aref 4 quad-array) (make-quad :a (elt corners 4) :b (elt corners 7)
					 :c (elt corners 3) :d (elt corners 0)
					 :n (normalize
					     (cross (vec- (elt corners 7) (elt corners 4))
						    (vec- (elt corners 3) (elt corners 4))))))
    (setf (aref 5 quad-array) (make-quad :a (elt corners 6) :b (elt corners 5)
					 :c (elt corners 1) :d (elt corners 2)
					 :n (normalize
					     (cross (vec- (elt corners 5) (elt corners 6))
						    (vec- (elt corners 1) (elt corners 6))))))
    quad-array))

;; VOL is the volume to render
;; CONTEXT is the rendering context for this volume
(defun volrender (vol context)
  ; IMAGE is the ouput image
  ; OBJTRANS transforms from object to viewing coordinates
  ; M transforms from viewing to object coordinates
  ; CORNERS is a list containing the transformed(by VOLUME-TRANS) object coordinates 
  ;   of the 8 corners of the dataset (i guess you could call these world coords)
  (let* ((image (make-array (list (volume-vol-height vol)
				  (volume-vol-width vol))))
	 (corners (mapcar #'(lambda (v) (m* v (volume-trans vol)))
			  (let ((maxx (volume-vol-max-width vol))
				(maxy (volume-vol-max-height vol)) 
				(maxz (volume-vol-max-depth vol)))
			    (list (make-vec4 :x 0.0d0 :y 0.0d0 :z 0.0d0 :w 1.0d0)
				  (make-vec4 :x maxx  :y 0.0d0 :z 0.0d0 :w 1.0d0)
				  (make-vec4 :x maxx  :y 0.0d0 :z maxz  :w 1.0d0)
				  (make-vec4 :x 0.0d0 :y 0.0d0 :z maxz  :w 1.0d0)
				  (make-vec4 :x 0.0d0 :y maxy  :z 0.0d0 :w 1.0d0)
				  (make-vec4 :x maxx  :y maxy  :z 0.0d0 :w 1.0d0)
				  (make-vec4 :x maxx  :y maxy  :z maxz  :w 1.0d0)
				  (make-vec4 :x 0.0d0 :y maxy  :z maxz  :w 1.0d0)))))
	 (objtrans (multiple-value-bind (minx maxx miny maxy minz maxz)
		       (get-projnorm-bounds vol corners)
		     (m* (volume-trans vol)
			 (proj-norm minx maxx miny maxy minz maxz))))
	 (m (inverse objtrans)))

    ; Just a little debugging info
    (format t "Corner points in viewspace: ~%")
    (mapcar #'print-vector corners)
    (multiple-value-bind (a b c d e f) (get-projnorm-bounds vol corners)
      (format t "X: ~A to ~A ~%Y: ~A to ~A ~%Z: ~A to ~A ~%" a b c d e f))
    (format t "Object to Viewspace: ~%")
    (print-matrix objtrans)
    (format t "Viewspace to Object: ~%")
    (print-matrix m)
    
    ; Optimizations for GET-VOL-VALUE
    (setf vol-width-fixnum vol-width)
    (setf vol-height-fixnum vol-height)

    ; Start rendering the image
    ; J is the row and I is the column of the output pixel
    ; (0,0) is the upper left-hand corner in screen/image coordinates
    ; SCREEN-TO-VIEW converts (I,J) into an (X,Y) coord on the projection plane
    ; A ray is then cast from this point, returning the final pixel value
    ; DZ is the step size in view-space
    (let ((vmw (truncate (volume-vol-max-width vol)))
	  (vmh (truncate (volume-vol-max-height vol)))
	  (vmw-df (volume-vol-max-width vol))
	  (vmh-df (volume-vol-max-height vol))
	  (starti 0)
	  (endi 0)
	  (startj 0)
	  (endj 0)
	  (viewspace-dz 1.0d0)
	  (dx 1.0d0)
	  (dy 1.0d0)
	  (dz 1.0d0))
      (declare (fixnum vmw vmh starti endi startj endj)
	       (double-float vmw-df vmh-df viewspace-dz dx dy dz))
      ;(setf starti 0)
      (setf endi vmw)
      ;(setf startj 0)
      (setf endj vmh)
      (multiple-value-bind (a b c d e f) (get-projnorm-bounds vol corners)
	(setf viewspace-dz (/ 2.0d0
			      (* (- f e) (volume-sampling-rate vol)))))
      (let* ((v1 (m* (make-vec4 :z 1.0d0 :w 1.0d0) m))
	     (v2 (m* (make-vec4 :z (- 1.0d0 viewspace-dz) :w 1.0d0) m)))
	(setf dx (- (vec4-x v2) (vec4-x v1)))
	(setf dy (- (vec4-y v2) (vec4-y v1)))
	(setf dz (- (vec4-z v2) (vec4-z v1))))
      (format t "DX=~A~% DY=~A~% DZ=~A~%" dx dy dz)
      
      (do ((j startj (+ j 1)))
	  ((> j endj))
	(declare (fixnum j))
	(do ((i starti (+ i 1)))
	    ((> i endi))
	  (declare (fixnum i))
	  
	  (multiple-value-bind (x y) (screen-to-view i j vmw vmh)
	    (setf (aref image j i) (cast-ray-slow vol context m x y dx dy dz viewspace-dz))))))
	    ;(setf (aref image j i) (cast-ray vol context m x y dx dy dz
					;(make-vec4 :x (float i 1.0d0) :y (float j 1.0d0) :z 255.0d0)
					;(make-vec4 :x (float i 1.0d0) :y (float j 1.0d0) :z 0.0d0)))))))
    
    (write-ppm "vol.ppm" image (volume-vol-width vol) (volume-vol-height vol))))
