#include <fcntl.h>
#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <stdlib.h>

#include "volume_io.h"

#define LENDIAN
#define DEBUG 0

/*+-------------------------------------------------------------------*/
/*+---------------+
  |   SwapBytes   |
  +---------------+*/ 
#ifdef  LENDIAN
void	SwapBytes(
   int	*intPtr,    /* pointer of words requiring byte swapping */
   unsigned int	sizeBytes   /* number of bytes in array */
){
   int				sizeInts, i;
   unsigned char	*bytePtr;

   sizeInts = sizeBytes / 4;
   for (i = 0; i < sizeInts; i++)
   {
      bytePtr = (unsigned char *) intPtr;
      *intPtr++ = ((bytePtr[0]<<24) | (bytePtr[1]<<16) | (bytePtr[2]<<8) | bytePtr[3]);
   }
}
#endif

int readVol(int input_fd, char *filename){
   int     l;
   int     totalDepth;
   int     bytesPerPixel;
   int     pixelsPerSlice;
   int     pixelsPerVolume;
   int     bytesPerSlice;

   enum	  errorCode;
   
   read(input_fd, (void *) &header_length, sizeof(int));
   read(input_fd, (void *) &vol_width, sizeof(int));
   read(input_fd, (void *) &vol_height, sizeof(int));
   read(input_fd, (void *) &vol_images, sizeof(int));
   read(input_fd, (void *) &pixel_depth, sizeof(int));
   read(input_fd, (void *) &index_bits, sizeof(int));
#ifdef  LENDIAN
   SwapBytes((int *) &header_length, sizeof(int));
   SwapBytes((int *) &vol_width, sizeof(int));
   SwapBytes((int *) &vol_height, sizeof(int));
   SwapBytes((int *) &vol_images, sizeof(int));
   SwapBytes((int *) &pixel_depth,sizeof(int));
   SwapBytes((int *) &index_bits, sizeof(int));
#endif
   pixel_depth -= index_bits;
   if (magic_number == VR_MAGIC_NUMBER){
      read( input_fd, (void *) &scale_x, sizeof(float));
      read( input_fd, (void *) &scale_y, sizeof(float));
      read( input_fd, (void *) &scale_z, sizeof(float));
      read( input_fd, (void *) &rot_x, sizeof(float));
      read( input_fd, (void *) &rot_y, sizeof(float));
      read( input_fd, (void *) &rot_z, sizeof(float));
#ifdef  LENDIAN
      SwapBytes ((int *) &scale_x, sizeof(float));
      SwapBytes ((int *) &scale_y, sizeof(float));
      SwapBytes ((int *) &scale_z, sizeof(float));
      SwapBytes ((int *) &rot_x, sizeof(float));
      SwapBytes ((int *) &rot_y, sizeof(float));
      SwapBytes ((int *) &rot_z, sizeof(float));
#endif
   }
   else{
      scale_x = (float) 1.0;
      scale_y = (float) 1.0;
      scale_z = (float) 1.0;
      rot_x = (float) 0.0;
      rot_y = (float) 0.0;
      rot_z = (float) 0.0;
   }


/*----- move the file pointer to the first voxel -----*/

   lseek(input_fd, header_length, SEEK_SET);        


/*----- first, let's figure out the pixel format -----*/

   totalDepth = pixel_depth + index_bits;

   switch (totalDepth){
      case 8:
         pixel_format = INTENSITY8;
         pixel_type = UNSIGNED_BYTE;
         break;

      case 16:
         if (index_bits == 0){
            pixel_format = INTENSITY16;
            pixel_type = UNSIGNED_SHORT;
         }
         else if (index_bits == 4){
            pixel_format = INDEX4_INTENSITY12;
            pixel_type = UNSIGNED_SHORT;
         }
         else if (index_bits == 8){
            pixel_format = INDEX8_INTENSITY8;
            pixel_type = UNSIGNED_SHORT;
         }
         else{
            printf (
                   "This combination of index (%d bits) & pixel (%d bits)\n",
                   index_bits, pixel_depth);
            printf ("depths is not supported\n");
            return (-1);
         }
         break;

      default:
         printf ("This combination of index (%d bits) & pixel (%d bits)\n",
                index_bits, pixel_depth);
         printf ("depths is not supported\n");
         return (-1);
         break;
   }


/*----- determine the size of the dataset -----*/

   if(pixel_type == UNSIGNED_BYTE)
      bytesPerPixel = sizeof (unsigned char);
   else 
      bytesPerPixel = sizeof (unsigned short) / 2;

   pixelsPerSlice = vol_width * vol_height;
   pixelsPerVolume = pixelsPerSlice * vol_images;
   bytesPerSlice = pixelsPerSlice * bytesPerPixel;


/*----- create buffer big enough for the volume -----*/

   rawdata = (unsigned char *) calloc (pixelsPerVolume, bytesPerPixel);
   if(rawdata == (unsigned char *) NULL){
      printf ("An error occurred allocating a volume buffer\n");
      exit (-1);
   }


/*----- read slices from the file and add to the volume -----*/

#if DEBUG
   printf ("bytesPerPixel = %d\n", bytesPerPixel);
   printf ("pixelsPerVolume = %d\n", pixelsPerVolume);
#endif

   if ((l = read(input_fd, rawdata, (bytesPerPixel * pixelsPerVolume)))
      != (bytesPerPixel * pixelsPerVolume))
   {
      printf ("Reading volume failed\n");
      close(input_fd);
      free(rawdata);
      return (-1);
   }


/*----- close file and free buffer -----*/

   close(input_fd);


/*----- print information about the loaded dataset -----*/
#if DEBUG
   printf ("\nFile: [ %s ]\n", filename);
   printf ("Image Width  = %d pixels\n", vol_width);
   printf ("Image Height = %d pixels\n", vol_height);
   printf ("Image Frames = %d pixels\n", vol_images);
   printf ("Pixel depth  = %d bits\n", pixel_depth);

   switch (pixel_format){
      case  INTENSITY8:
         printf ("Pixel format = INTENSITY8\n");
         break;

      case  INTENSITY16:
         printf ("Pixel format = INTENSITY16\n");
         break;

      case  INDEX4_INTENSITY12:
         printf ("Pixel format = INDEX4_INTENSITY12\n");
         break;

      case  INDEX8_INTENSITY8:
         printf ("Pixel format = INDEX8_INTENSITY8\n");
         break;
   }

   if(pixel_type == UNSIGNED_BYTE )
      printf ("Image type  = UNSIGNED_BYTE\n");
   else
      printf ("Image type  = UNSIGNED_SHORT\n");

   if(manufacturer[0])
      printf ("Original manufacturer is %s\n", manufacturer);
   if(original_filename[0])
      printf ("Original file name is %s\n", original_filename);
#endif /* DEBUG */
   
   return 1;
}

int read_volume(char *filename){
   int     input_fd;

   /*----- open the volume file -----*/
   
   if ((input_fd = open(filename, O_RDONLY)) == -1){
      printf ("Can't open volume file: %s\n", filename);
      return (-1);
   }

   /*----- read header info from the volume file -----*/

   read(input_fd, &magic_number, sizeof(int));
#if DEBUG
   printf("Magic Number is %x\n", magic_number);
#endif   
   
   return readVol(input_fd, filename);

}

#if 0
int main(int argc, char **argv){
    
  if(argc < 2) return -1;
  read_volume(argv[1]);
  
  return;
}
#endif






